/*
 * PUBLIC DOMAIN PCCTS-BASED C++ GRAMMAR (cplusplus.g, stat.g, expr.g)
 *
 * Authors: Sumana Srinivasan, NeXT Inc.;            sumana_srinivasan@next.com
 *          Terence Parr, Parr Research Corporation; parrt@parr-research.com
 *          Russell Quong, Purdue University;        quong@ecn.purdue.edu
 *
 * VERSION 1.1
 *
 * SOFTWARE RIGHTS
 *
 * This file is a part of the ANTLR-based C++ grammar and is free
 * software.  We do not reserve any LEGAL rights to its use or
 * distribution, but you may NOT claim ownership or authorship of this
 * grammar or support code.  An individual or company may otherwise do
 * whatever they wish with the grammar distributed herewith including the
 * incorporation of the grammar or the output generated by ANTLR into
 * commerical software.  You may redistribute in source or binary form
 * without payment of royalties to us as long as this header remains
 * in all source distributions.
 *
 * We encourage users to develop parsers/tools using this grammar.
 * In return, we ask that credit is given to us for developing this
 * grammar.  By "credit", we mean that if you incorporate our grammar or
 * the generated code into one of your programs (commercial product,
 * research project, or otherwise) that you acknowledge this fact in the
 * documentation, research report, etc....  In addition, you should say nice
 * things about us at every opportunity.
 *
 * As long as these guidelines are kept, we expect to continue enhancing
 * this grammar.  Feel free to send us enhancements, fixes, bug reports,
 * suggestions, or general words of encouragement at parrt@parr-research.com.
 * 
 * NeXT Computer Inc.
 * 900 Chesapeake Dr.
 * Redwood City, CA 94555
 * 12/02/1994
 * 
 * Restructured for public consumption by Terence Parr late February, 1995.
 *
 * Requires PCCTS 1.32b4 or higher to get past ANTLR. 
 * 
 * DISCLAIMER: we make no guarantees that this grammar works, makes sense,
 *             or can be used to do anything useful.
 */
#include <iostream.h>
#include "tokens.h"
#include "CPPParser.h"
typedef ANTLRCommonToken ANTLRToken;
#include "CPPSymbol.h"

// Shorthand for a string of (qualifiedItemIs()==xxx||...)
int CPPParser::qualifiedItemIsOneOf(
   QualifiedItem qiFlags,   // Ored combination of flags
   int lookahead_offset
)
{ 
   QualifiedItem qi = qualifiedItemIs(lookahead_offset); 
   if (doTracing)
   {
      fprintf(
         stdout, 
         "qualifiedItemIsOneOf(0x%x,%d) returns %d (item is 0x%x)\n", 
         qiFlags,
         lookahead_offset, 
         ((qi & qiFlags) != 0),
         qi
      );
   }
   return (qi & qiFlags) != 0; 
}

// This is an important function, but will be replaced with
// an enhanced predicate in the future, once predicates
// and/or predicate guards can contain loops.
//
// Scan past the ::T::B:: to see what lies beyond.
// Return qiType if the qualified item can pose as type name.
// Note that T::T is NOT a type; it is a constructor.  Also,
// class T { ... T...} yields the enclosed T as a ctor.  This
// is probably a type as I separate out the constructor defs/decls,
// I want it consistent with T::T.
//
// In the below examples, I use A,B,T and example types, and
// a,b as example ids.
// In the below examples, any A or B may be a
// qualified template, i.e.,  A<...>
//
// T::T outside of class T yields qiCtor.
// T<...>::T outside of class T yields qiCtor.
// T inside of class T {...} yields qiCtor.
// T, ::T, A::T outside of class T yields qiType.
// a, ::a,  A::B::a yields qiId
// a::b yields qiInvalid
// ::operator, operator, A::B::operator yield qiOPerator
// A::*, A::B::* yield qiPtrMember
// ::*, * yields qiInvalid
// ::~T, ~T, A::~T yield qiDtor
// ~a, ~A::a, A::~T::, ~T:: yield qiInvalid

CPPParser::QualifiedItem 
CPPParser::qualifiedItemIs(
   int lookahead_offset
)
{
   int k = lookahead_offset + 1;
   int final_type_idx = 0;

   // Skip leading "::"
   if ( LT(k)->getType() == SCOPE )
   {
      k++;
   }

   // Skip sequences of T:: or T<...>::
   while (LT(k)->getType() == ID && isTypeName(LT(k)->getText()))
   {
      // If this type is the same as the last type, then ctor
      if (
         final_type_idx != 0 && 
         strcmp(LT(final_type_idx)->getText(), LT(k)->getText()) == 0
      )
      {
         // Like T::T
         // As an extra check, do not allow T::T::
         if (LT(k+1)->getType() == SCOPE)
         {
            return qiInvalid;
         } else {
            return qiCtor;
         }
      }

      // Record this as the most recent type seen in the series
      final_type_idx = k;

      // Skip this token
      k++;

      // Skip over any template qualifiers <...>
      // I believe that "T<..." cannot be anything valid but a template
      if (LT(k)->getType() == LESSTHAN)
      {
         if (!skipTemplateQualifiers(k))
         {
            return qiInvalid;
         }
         // k has been updated to token following <...>
      }

      if (LT(k)->getType() == SCOPE)
      {
         // Skip the "::" and keep going
         k++;
      } else {
         // Series terminated -- last ID in the sequence was a type
         // Return ctor if last type is in containing class
         // We already checked for T::T inside loop
         if ( strcmp(enclosingClass, LT(final_type_idx)->getText())==0 ) {
            // Like class T { T()
            return qiCtor;
         } else {
            return qiType;
         }
      }
   }

   // LT(k) is not an ID, or it is an ID but not a typename.
   switch (LT(k)->getType())
   {
      case ID:
         // ID but not a typename
         // Do not allow id::
         if (LT(k+1)->getType() == SCOPE)
         {
            return qiInvalid;
         }
         if ( strcmp(enclosingClass,LT(k)->getText())==0 ) {
            // Like class T { T()
            return qiCtor;
         } else {
            return qiID;
         }
      case TILDE:
         // check for dtor
         if (LT(k+1)->getType() == ID && 
            isTypeName(LT(k+1)->getText()) &&
            LT(k+2)->getType() != SCOPE
         )
         {
            // Like ~B or A::B::~B
            // Also (incorrectly?) matches ::~A.
            return qiDtor;
         } else {
            // ~a or ~A::a is qiInvalid
            return qiInvalid;
         }
         break;
      case STAR:
         // Like A::*
         // Do not allow * or ::*
         if (final_type_idx == 0)
         {
            // Haven't seen a type yet
            return qiInvalid;
         } else {
            return qiPtrMember;
         }
      case OPERATOR:
         // Like A::operator, ::operator, or operator
         return qiOperator;
      default:
         // Something that neither starts with :: or ID, or
         // a :: not followed by ID, operator, ~, or *
         return qiInvalid;
   }
}


// Skip over <...>.  This correctly handles nested <> and (), e.g:
//    <T>
//    < (i>3) >
//    < T2<...> >
// but not
//    < i>3 >
//
// On input, kInOut is the index of the "<"
// On output, if the return is true, then 
//                kInOut is the index of the token after ">"
//            else
//                kInOut is unchanged
int CPPParser::skipTemplateQualifiers(int& kInOut)
{
   // Start after "<"
   int k = kInOut + 1;

   while ( LT(k)->getType() != GREATERTHAN ) // scan to end of <...>
   {
      switch ( LT(k)->getType())
      {
         case Eof:
            return 0;
         case LESSTHAN:
            if (!skipTemplateQualifiers(k))
            {
               return 0;
            }
            break;
         case LPAREN:
            if (!skipNestedParens(k))
            {
               return 0;
            }
            break;
         default:
            k++;     // skip everything else
            break;
      }
      if (k > MaxTemplateTokenScan )
      {
         return 0;
      }
   }

   // Update output argument to point past ">"
   kInOut = k + 1;
   return 1;
}

// Skip over (...).  This correctly handles nested (), e.g:
//    (i>3, (i>5))
//
// On input, kInOut is the index of the "("
// On output, if the return is true, then 
//                kInOut is the index of the token after ")"
//            else
//                kInOut is unchanged
int CPPParser::skipNestedParens(int& kInOut)
{
   // Start after "("
   int k = kInOut + 1;

   while ( LT(k)->getType() != RPAREN )   // scan to end of (...)
   {
      switch ( LT(k)->getType())
      {
         case Eof:
            return 0;
         case LPAREN:
            if (!skipNestedParens(k))
            {
               return 0;
            }
            break;
         default:
            k++;     // skip everything else
            break;
      }
      if (k > MaxTemplateTokenScan )
      {
         return 0;
      }
   }

   // Update output argument to point past ")"
   kInOut = k + 1;
   return 1;
}

// Return true if "::blah" or "fu::bar<args>::..." found.
int CPPParser::scopedItem(int k)
{
   return ( 
      LT(k)->getType()==SCOPE ||
      (LT(k)->getType()==ID && !finalQualifier(k))
   );
}

// Return true if ID<...> or ID is last item in qualified item list.
// Return false if LT(k) is not an ID.
// ID must be a type to check for ID<...>,
// or else we would get confused by "i<3"
int CPPParser::finalQualifier(int k)
{
   if (LT(k)->getType()==ID)
   {
      if (isTypeName(LT(k)->getText()) && LT(k+1)->getType()==LESSTHAN)
      {
         // Starts with "T<".  Skip <...>
         k++;
         skipTemplateQualifiers(k);
      } else {
         // skip ID;
         k++;
      }
      return (LT(k)->getType() != SCOPE );
   } else {
      // not an ID
      return 0;
   }
}


/*
 * Return true if 's' can pose as a type name
 */
int CPPParser::
isTypeName(char *s)
{
   CPPSymbol *cs = (CPPSymbol *) symbols->lookup(s);
   if ( cs==NULL ) return 0;
   if ( cs->getType()==CPPSymbol::otTypedef ||
       cs->getType()==CPPSymbol::otClass ||
       cs->getType()==CPPSymbol::otEnum ||
       cs->getType()==CPPSymbol::otUnion ||
       cs->getType()==CPPSymbol::otStruct ) return 1;
   return 0;
}

/*
 * Return true if 's' is a class name (or a struct which is a class
 * with all members public).
 */
int CPPParser::
isClassName(char *s)
{
   CPPSymbol *cs = (CPPSymbol *) symbols->lookup(s);
   if ( cs==NULL ) return 0;
   if ( cs->getType()==CPPSymbol::otClass ||
       cs->getType()==CPPSymbol::otStruct ||
       cs->getType()==CPPSymbol::otUnion ) return 1;
   return 0;
}

void CPPParser::
beginDeclaration()
{
}

void CPPParser::
endDeclaration()
{
}

void CPPParser::
beginFunctionDefinition()
{
   functionDefinition = 1;
}

void CPPParser::
endFunctionDefinition()
{
// fprintf(stdout, "parameter scope(%d):\n",symbols->getCurrentScopeIndex());
   // Remove parameter scope
   symbols->dumpScope(stdout);
   symbols->removeScope();
   symbols->restoreScope();
   functionDefinition = 0;
}

void CPPParser::
beginConstructorDefinition()
{
   functionDefinition = 1;
}

void CPPParser::
endConstructorDefinition()
{
// fprintf(stdout, "ctor parameter scope(%d):\n",symbols->getCurrentScopeIndex());
   symbols->dumpScope(stdout);
   symbols->removeScope();
   symbols->restoreScope();
   functionDefinition = 0;
}

void CPPParser::
beginConstructorDeclaration(char *ctor)
{
}

void CPPParser::
endConstructorDeclaration()
{
}

void CPPParser::
beginDestructorDefinition()
{
   functionDefinition = 1;
}

void CPPParser::
endDestructorDefinition()
{
// fprintf(stdout, "parameter scope(%d):\n",symbols->getCurrentScopeIndex());
   symbols->dumpScope(stdout);
   symbols->removeScope();
   symbols->restoreScope();
   functionDefinition = 0;
}

void CPPParser::
beginDestructorDeclaration(char *dtor)
{
}

void CPPParser::
endDestructorDeclaration()
{
}

void CPPParser::
beginParameterDeclaration()
{
}

void CPPParser::
beginFieldDeclaration()
{
}

void CPPParser::
declarationSpecifier(StorageClass sc,
                TypeQualifier tq,
                TypeSpecifier ts)
{
   _sc = sc;
   _tq = tq;
   _ts = ts;
}

void CPPParser::
declaratorPointerTo()
{
}

/* Symbols from declarators are added to the symbol table here.  Note
 * that we only care about types for parsing (with semantic predicates)
 * and we don't track variable vs function.  A real C++ front-end would
 * have to, of course.  The symbol is added to whatever the current
 * scope is in the symbol table.
 */
void CPPParser::
declaratorID(char *id)
{
   CPPSymbol *c, *ext_c;
   if ( _sc==scTYPEDEF )
   {
      c = new CPPSymbol(id, CPPSymbol::otTypedef);
      ext_c = new CPPSymbol(id, CPPSymbol::otTypedef);
      if ( ext_c==NULL ) panic("can't alloc CPPSymbol");
      if ( symbols->getCurrentScopeIndex()!=externalScope )
         symbols->defineInScope(id, ext_c, externalScope);
   }
   else c = new CPPSymbol(id);
   if ( c==NULL ) panic("can't alloc CPPSymbol");
   symbols->define(id, c);
}

void CPPParser::
declaratorArray()
{
}

void CPPParser::
declaratorParameterList(int def)
{
   symbols->saveScope();
// fprintf(stdout, "entering def parm scope %d\n", symbols->getCurrentScopeIndex());
}

void CPPParser::
declaratorEndParameterList(int def)
{
   if ( !def )
   {
      /* this is for a decl only...remove immediately */
//    fprintf(stdout, "parameter scope(%d):\n",symbols->getCurrentScopeIndex());
      symbols->dumpScope(stdout);
      symbols->removeScope();
      symbols->restoreScope();
   }
}

void CPPParser::
functionParameterList()
{
   symbols->saveScope();
// fprintf(stdout, "entering func parm scope %d\n",
//       symbols->getCurrentScopeIndex());
}

void CPPParser::
functionEndParameterList()
{
   /* The end function thing removes the parameters from scope */
}

void CPPParser::
enterNewLocalScope()
{
   symbols->saveScope();
// fprintf(stdout, "entering local scope (%d):\n", symbols->getCurrentScopeIndex());
}

void CPPParser::
exitLocalScope()
{
   fprintf(stdout, "exiting local scope (%d):\n", symbols->getCurrentScopeIndex());
   symbols->dumpScope(stdout);
   symbols->removeScope();
   symbols->restoreScope();
}

void CPPParser::
enterExternalScope()
{
}

void CPPParser::
exitExternalScope()
{
   fprintf(stdout, "global scope:\n");
   symbols->dumpScope(stdout);
   symbols->removeScope();
}

void CPPParser::
classForwardDeclaration(TypeSpecifier ts, DeclSpecifier ds, char *tag)
{
   CPPSymbol *c = NULL;

   // if already in symbol table as a class, don't add
   // of course, this is incorrect as you can rename
   // classes by entering a new scope, but our limited
   // example basically keeps all types globally visible.
   if ( symbols->lookup(tag)!=NULL ) return;

   switch ( ts )
   {
     case tsSTRUCT :
      c = new CPPSymbol(tag, CPPSymbol::otStruct);
      break;
     case tsUNION :
      c = new CPPSymbol(tag, CPPSymbol::otUnion);
      break;
     case tsCLASS :
      c = new CPPSymbol(tag, CPPSymbol::otClass);
      break;
   }
   if ( c==NULL ) panic("can't alloc CPPSymbol");
   symbols->define(tag, c);

   // If it's a friend class forward decl, put in global scope also.
   if ( ds==dsFRIEND )
   {
      CPPSymbol *ext_c = new CPPSymbol(tag, CPPSymbol::otClass);
      if ( ext_c==NULL ) panic("can't alloc CPPSymbol");
      if ( symbols->getCurrentScopeIndex()!=externalScope )
         symbols->defineInScope(tag, ext_c, externalScope);
   }
}

void CPPParser::
beginClassDefinition(TypeSpecifier ts, char *tag)
{
   CPPSymbol *c, *ext_c;

   // if already in symbol table as a class, don't add
   // of course, this is incorrect as you can rename
   // classes by entering a new scope, but our limited
   // example basically keeps all types globally visible.
   if ( symbols->lookup(tag)!=NULL )
   {
      symbols->saveScope();   // still have to use scope to collect members
      return;
   }

   switch ( ts )
   {
     case tsSTRUCT :
      c = new CPPSymbol(tag, CPPSymbol::otStruct);
      ext_c = new CPPSymbol(tag, CPPSymbol::otStruct);
      break;
     case tsUNION :
      c = new CPPSymbol(tag, CPPSymbol::otUnion);
      ext_c = new CPPSymbol(tag, CPPSymbol::otUnion);
      break;
     case tsCLASS :
      c = new CPPSymbol(tag, CPPSymbol::otClass);
      ext_c = new CPPSymbol(tag, CPPSymbol::otClass);
      break;
   }
   if ( c==NULL ) panic("can't alloc CPPSymbol");
   if ( ext_c==NULL ) panic("can't alloc CPPSymbol");

   // this one lives only as long as the end '}' of an enclosing class def
   // I don't think I'll use this one to do the symbol look ups
   // I'll prepend the current qualifierPrefix before looking
   // things up.
   symbols->define(tag, c);
   strcat(qualifierPrefix, tag);
   strcat(qualifierPrefix, "::");

   // add all member type symbols into the global scope (not correct, but
   // will work for most code).
   // This symbol lives until the end of the file
   if ( symbols->getCurrentScopeIndex()!=externalScope )
      symbols->defineInScope(tag, ext_c, externalScope);

   symbols->saveScope();   // use the scope to collect list of fields
}

void CPPParser::
endClassDefinition()
{
   fprintf(stdout, "list of fields:\n");
   symbols->dumpScope(stdout);
   symbols->removeScope();
   symbols->restoreScope();

   // remove final T:: from A::B::C::T::
   // upon backing out of last class, qualifierPrefix is set to ""
   char *p = &(qualifierPrefix[strlen(qualifierPrefix)-3]);
   while ( p > &(qualifierPrefix[0]) && *p!=':' ) {p--;}
   if ( p > &(qualifierPrefix[0]) ) p++;
   *p = '\0';
}

void CPPParser::
enumElement(char *e)
{
}

void CPPParser::
beginEnumDefinition(char *e)
{
   // add all enum tags into the global scope (not correct, but
   // will work for most code).
   // This symbol lives until the end of the file
   CPPSymbol *ext_e = new CPPSymbol(e, CPPSymbol::otEnum);
   symbols->defineInScope(e, ext_e, externalScope);
}

void CPPParser::
endEnumDefinition()
{
}

void CPPParser::
templateTypeParameter(char *t)
{
   // Add as a temporary typedef that we'll remove at the end of the template
   CPPSymbol *a = new CPPSymbol(t, CPPSymbol::otTypedef);
   symbols->defineInScope(t, a, templateParameterScope);
}

void CPPParser::
beginTemplateDefinition()
{
}

void CPPParser::
endTemplateDefinition()
{
   // remove elements from the scope, but don't destroy it; we'll need it again
   fprintf(stdout, "list of template params:\n");
   symbols->dumpScope(stdout,templateParameterScope);
   symbols->removeScope(templateParameterScope);
}

void CPPParser::
beginTemplateParameterList()
{
}

void CPPParser::
endTemplateParameterList()
{
}

void CPPParser::
exceptionBeginHandler()
{
}

void CPPParser::
exceptionEndHandler()
{
   // remove parm elements from the hendler scope
   fprintf(stdout, "list of handler params:\n");
   symbols->dumpScope(stdout);
   symbols->removeScope();
   symbols->restoreScope();
}

void CPPParser::
panic(char *err)
{
   fprintf(stdout, "CPPParser: %s\n", err);
   exit(-1);
}

void CPPParser::
tracein(char *r)
{
   if ( !doTracing ) return;
   for (int i=1; i<=traceIndentLevel; i++) fprintf(stdout, " ");
   traceIndentLevel++;
   fprintf(stdout, "enter %s('%s %s')%s\n",
      r,
      LT(1)->getText(),
      LT(2)->getText(),
      guessing?" [guessing]":""
   );
   fflush(stdout);
}

void CPPParser::
traceout(char *r)
{
   if ( !doTracing ) return;
   traceIndentLevel--;
   for (int i=1; i<=traceIndentLevel; i++) fprintf(stdout, " ");
   fprintf(stdout, "exit %s('%s %s')%s\n",
      r,
      LT(1)->getText(),
      LT(2)->getText(),
      guessing?" [guessing]":""
   );
   fflush(stdout);
}

char *CPPParser::
qualified(char *s)
{
   static char q[CPPParser_MaxQualifiedItemSize+1];
   strcpy(q, qualifierPrefix);
   strcat(q, s);
   return q;
}


